/*
** frame.c
**
** This program demonstrates the construction of a very simple
** "container" on a system with a Linux 4.4 kernel. Other than
** constructing the "container", it does no useful work.
**
** I wrote this program to explore some variations in the steps
** necessary to construct a "container". Conditional compilation
** (by defining or not defining the "ALT" macro) permits A/B
** testing of variations.
**
** Copyright (C) Lew Pitcher (2023)
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License along
** with this program; if not, write to the Free Software Foundation, Inc.,
** 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
**
** Lew Pitcher <lew.pitcher@digitalfreehold.ca>
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <dirent.h>
#include <sched.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mount.h>
#include <sys/wait.h>

/* pivot_root() function prototype not supplied by standard headers */
extern int pivot_root(const char *new_root, const char *put_old);

#define		NENTRIES(x)	(sizeof(x)/sizeof(*x))

static int container(char *path);
static int assumeRoot(uid_t oldUid, gid_t oldGid);
static int toyInit(void);

static int utInitDirectory(char *path);
static int utRmContents(char *path);

int main(int argc, char *argv[])
{
  int success = 0;	/* assume failure */

#ifndef ALT
  puts("===== MAINLINE flow =====");
#else
  puts("===== ALTERNATE flow =====");
#endif

  umask(0);

  if (argc == 2)
  {
    if (success = utInitDirectory(argv[1]))
      success = container(argv[1]);
  }
  else fprintf(stderr,"Usage: %s path_to_work_directory\n",argv[0]);

  return success ? EXIT_SUCCESS : EXIT_FAILURE;
}

/*
** container() will direct the construction of a minimal "container"
*/
static int container(char *path)
{
  int progress = 1;	/* assume success */
  uid_t	host_uid = geteuid();
  gid_t	host_gid = getegid();


  printf("unshare(CLONE_NEWUSER|CLONE_NEWNS|CLONE_NEWPID|CLONE_NEWNET)\n  returned ");
  if (unshare(CLONE_NEWUSER|CLONE_NEWNS|CLONE_NEWPID|CLONE_NEWNET) == 0)
  {
    puts("SUCCESS\n");

    progress = assumeRoot(host_uid,host_gid);

    /*
    ** Change the propogation type of current root directory, and 
    ** any mounts below it, to PRIVATE so that mount and unmount
    ** events do not propagate into or out of them.
    */
    if (progress)
    {
      printf("mount(NULL,\"/\",NULL,MS_REC,MS_PRIVATE,NULL)\n  returned ");
      if (mount(NULL,"/",NULL,MS_REC|MS_PRIVATE,NULL))
      {
	printf("\"%s\"\n",strerror(errno));
	progress = 0;
      }
      else puts("SUCCESS");
      puts("");
    }

    /*
    ** Make our target directory into a mount point, in preparation for
    ** a pivot_root() call. This satisfies two conditions of pivot_root():
    **   1) the "new root" (our cwd) must not be in the same filesystem
    **      as the current root, and
    **   2) the "new root" (our cwd) must be a path to a mount point.
    */
    if (progress)
    {
      printf("mount(\"%s\",\"%s\",NULL,MS_BIND|MS_NOSUID,NULL)\n  returned ",path,path);
      if (mount(path,path,NULL,MS_BIND|MS_NOSUID,NULL))
      {
	printf("\"%s\"\n",strerror(errno));
	progress = 0;
      }
      else puts("SUCCESS");
      puts("");
    }

    /*
    ** Now, move into the directory that would be root
    */
    if (progress)
    {
      printf("chdir(\"%s\")\n  returned ",path);
      if (chdir(path))
      {
	printf("\"%s\"\n",strerror(errno));
	progress  = 0;
      }
      else puts("SUCCESS");
      puts("");
    }

    /*
    ** pivot our root directory
    ** the old root directory goes to ./tmp
    ** and cwd becomes our new root directory
    */
    if (progress)
    {
      printf("pivot_root(\".\",\"tmp\")\n  returned ");
      if (pivot_root(".","tmp"))
      {
	printf("\"%s\"\n",strerror(errno));
	progress = 0;
      }
      else puts("SUCCESS");
      puts("");
    }

    /*
    ** pivot_root(2) recommends chdir("/") after pivot_root() call
    */
    if (progress)
    {
      printf("chdir(\"/\")\n  returned ");
      if (chdir("/"))
      {
	printf("\"%s\"\n",strerror(errno));
	progress  = 0;
      }
      else puts("SUCCESS");
      puts("");
    }

    /* 
    ** launch init, finish our mounts in that process
    */
    if (progress)
    {
      switch(fork())
      {
	case 0:
	  toyInit();
	  break;

	case -1:
	  fprintf(stderr,"ERROR: fork()\n  returned \"%s\"\n",strerror(errno));
	  progress = 0;
	  break;

	default:
	  {
	    int status;

	    wait(&status);
	    if (!WIFEXITED(status) || (WEXITSTATUS(status) != 0))
	    {
	      fprintf(stderr,"ERROR: toy init failed\n");
	      progress = 0;
	    }
	  }
	  break;
      }
    }
  }
  else
  {
    printf("\"%s\"\n\n",strerror(errno));
    progress = 0;
  }

  return progress;
}

/*
** assumeRoot(oldUid, oldGid)
**
** Update the process entry in /proc to map our outside-the-box 
** uid and gid to the inside-the-box root user uid and gid,
** matching the effects of our new CLONE_NEWUSER capabilities.
**
** Accepts:	uid_t oldUid - our outside-the-box UID,
**		gid_t oldGid - our outside-the-box GID
**
** Returns: 	int 1  - success
**	    	int 0  - failure
*/
static int assumeRoot(uid_t oldUid, gid_t oldGid)
{
  int	progress = 1,	/* assume success */ 
	fd;
  FILE *map;

  /*
  ** Map our uid into the container's root uid (UID 0)
  */
  if (map = fopen("/proc/self/uid_map","w"))
  {
    fprintf(map,"0 %lu 1\n",(unsigned long)oldUid);
    fclose(map);
  }
  else
  {
    fprintf(stderr,"ERROR: fopen(\"/proc/self/uid_map\") \n  returned %s\n",strerror(errno));
    progress = 0;
  }

  /*
  ** Some kernels require /proc/self/setgroups set to "deny"
  ** before we can remap our GID
  */
  if (progress  && ((fd = open("/proc/self/setgroups",O_WRONLY)) >= 0))
  {
    write(fd,"deny",4);
    close(fd);
  }

  if (progress)
  {
    /*
    ** Map our gid into the container's root gid (GID 0)
    */
    if (map = fopen("/proc/self/gid_map","w"))
    {
      fprintf(map,"0 %lu 1\n",(unsigned long)oldGid);
      fclose(map);
    }
    else
    {
      fprintf(stderr,"ERROR: fopen(\"/proc/self/gid_map\") \n  returned %s\n",strerror(errno));
      progress = 0;
    }
  }

  return progress;
}

/*
** toyInit() will perform the functions of our container's init process
*/
static int toyInit(void)
{
  int progress = 1;

  if (progress)
  {
    printf("INIT: mount(\"none\",\"/proc\",\"proc\",MS_REC,NULL)\n  returned ");
    if (mount("none","/proc","proc",MS_REC,NULL))
    {
      printf("\"%s\"\n",strerror(errno));
      progress = 0;
    }
    else puts("SUCCESS");
    puts("");
  }

  if (progress)
  {
    printf("INIT: mount(\"none\",\"/sys\",\"sysfs\",0,NULL)\n  returned ");
    if (mount("none","/sys","sysfs",0,NULL))
    {
      printf("\"%s\"\n",strerror(errno));
      progress = 0;
    }
    else puts("SUCCESS");
    puts("");
  }

  /*
  ** remove the old root directory from our mount tree
  ** Note: also releases /tmp/proc and /tmp/sys
  */
  if (progress)
  {
    printf("INIT: umount2(\"/tmp\",MNT_DETACH)\n  returned ");
    if (umount2("/tmp",MNT_DETACH))
    {
      printf("\"%s\"\n",strerror(errno));
      progress = 0;
    }
    else puts("SUCCESS");
    puts("");
  }

  /*
  ** >>>>>>>>>> Give ourselves a tempfs for our /tmp directory <<<<<<<<<<
  */
  if (progress)
  {
    printf("INIT: mount(\"none\",\"/tmp\",\"tmpfs\",MS_NOSUID|MS_NODEV|MS_NOEXEC|MS_NOATIME,...)\n  returned ");
    if (mount("none","/tmp","tmpfs",MS_NOSUID|MS_NODEV|MS_NOEXEC|MS_NOATIME,"size=16m,nr_inodes=4k,mode=01777"))
    {
      printf("\"%s\"\n",strerror(errno));
      progress = 0;
    }
    else puts("SUCCESS");
    puts("");
  }

  if (progress)
  {
    puts("Note:\tAt this point, we have established a container that would");
    puts("\tnormally do some useful work like launching daemons and");
    puts("\tsuch. However, as THIS program exists only to sketch out");
    puts("\tthe steps needed to get this far, this program has now");
    puts("\taccomplished it's goal, and will terminate.");
  }

  _exit(progress?EXIT_SUCCESS:EXIT_FAILURE);
}

/*******************************************************************************
******************************** UTILITY FUNCTIONS *****************************
*******************************************************************************/

/*
** utInitDirectory() constructs the target directory and it's requisite subdirectories
*/
static int utInitDirectory(char *path)
{
  int progress = 1;	/* default to SUCCESS */
  struct {
    char *path;
    mode_t mode;
  } directory[] = {{"proc",0555},
		   {"sys",0555},
		   {"tmp",01777}
		  };
  struct stat buf;

  /*
  ** clear or create the requested directory
  */
  if (stat(path,&buf) == 0)
  {
    if (S_ISDIR(buf.st_mode))
    {
      progress = utRmContents(path);	/* clean up test directory prior to test */
    }
    else
    {
      fprintf(stderr,"ERROR: \"%s\" is not a directory\n",path);
      progress = 0;
    }
  }
  else
  {
    if (errno == ENOENT)
    {
      if (mkdir(path,0777) != 0)
      {
	fprintf(stderr,"ERROR: can't mkdir(\"%s\",0777): %s\n",path,strerror(errno));
	progress = 0;
      }
      else printf("PREP: Created directory \"%s\"\n",path);
    }
    else
    {
      fprintf(stderr,"ERROR: can't stat(\"%s\"): %s\n",path,strerror(errno));
      progress = 0;
    }
  }

  /*
  ** fill directory with required subdirectories
  */
  for (int index = 0; progress && (index < NENTRIES(directory)); ++index)
  {
    char *subpath;

    if (subpath = malloc(strlen(path)+strlen(directory[index].path)+2))
    {
      sprintf(subpath,"%s/%s",path,directory[index].path);

      if (mkdir(subpath,directory[index].mode))
      {
	progress = 0;
	fprintf(stderr,"ERROR: mkdir(\"%s\",0%4o) returned \"%s\"\n",
		subpath,directory[index].mode,strerror(errno));
      }
      else printf("PREP: Created directory \"%s\"\n",subpath);
      free(subpath);
    }
    else
    {
      progress = 0;
      fprintf(stderr,
	      "ERROR: Can't allocate memory to build %s/%s subdirectory path\n",
	      path,
	      directory[index].path);
    }
  }
  puts("");

  return progress;
}

/*
** utRmContents() removes the contents of a specified directory
*/
static int utRmContents(char *path)
{
  int progress = 1;
  DIR *dir;

  if (dir = opendir(path))
  {
    struct dirent *dirent;

    while (progress && (dirent = readdir(dir)))
    {
      if (strcmp(dirent->d_name,".") && strcmp(dirent->d_name,".."))
      {
	char *subpath;

	if (subpath = malloc(strlen(path)+strlen(dirent->d_name)+2))
	{
	  struct stat sp_buf;

	  sprintf(subpath,"%s/%s",path,dirent->d_name);
	  if (stat(subpath,&sp_buf) == 0)
	  {
	    if (S_ISDIR(sp_buf.st_mode))
	    {
	      if (progress = utRmContents(subpath))
		if (rmdir(subpath) == -1)
		{
		  fprintf(stderr,"ERROR: rmdir(\"%s\")\n  returned %s\n",subpath,strerror(errno));
		  progress = 0;
		}
		else printf("PREP: Deleted directory \"%s\"\n",subpath);
	    }
	    else
	    {
	      if (unlink(subpath) == -1)
	      {
		fprintf(stderr,"ERROR: unlink(\"%s\")\n  returned %s\n",subpath,strerror(errno));
		progress = 0;
	      }
	    }
	  }
	  free(subpath);
	}
	else
	{
	  fprintf(stderr,"ERROR: Can't allocate space for recursive delete\n");
	  progress = 0;
	}
      }
    }
    closedir(dir);
  }
  else progress = 0;

  if (progress) printf("PREP: Deleted contents of directory \"%s\"\n",path);

  return progress;
}
